#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""example.py"""
# pylint: disable=C0116
# pylint: disable=W0613

from datetime import timedelta
from typing import Dict, Any
import typing
from restate import Service, Context

from . import awakable_holder

test_utils = Service("TestUtilsService")

@test_utils.handler()
async def echo(context: Context, input: str) -> str:
    return input

@test_utils.handler(name="uppercaseEcho")
async def uppercase_echo(context: Context, input: str) -> str:
    return input.upper()

@test_utils.handler(name="echoHeaders")
async def echo_headers(context: Context) -> Dict[str, str]:
    return context.request().headers

@test_utils.handler(name="createAwakeableAndAwaitIt")
async def create_awakeable_and_await_it(context: Context, req: Dict[str, Any]) -> Dict[str, Any]:
    name, awakeable = context.awakeable()

    await context.object_call(awakable_holder.hold, key=req["awakeableKey"], arg=name)

    if "awaitTimeout" not in req:
        return {"type": "result", "value": await awakeable}

    timeout = context.sleep(timedelta(milliseconds=int(req["awaitTimeout"])))
    raise NotImplementedError()

@test_utils.handler(name="sleepConcurrently")
async def sleep_concurrently(context: Context, millis_duration: typing.List[int]) -> None:
    timers = [context.sleep(timedelta(milliseconds=duration)) for duration in millis_duration]

    for timer in timers:
        await timer


@test_utils.handler(name="countExecutedSideEffects")
async def count_executed_side_effects(context: Context, increments: int) -> int:
    invoked_side_effects = 0

    def effect():
        nonlocal invoked_side_effects
        invoked_side_effects += 1

    for _ in range(increments):
        await context.run("count", effect)

    return invoked_side_effects
    



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
This module contains the definition of the Handler class,
which is used to define the handlers for the services.
"""

from dataclasses import dataclass
from typing import Any, Callable, Awaitable, Generic, Literal, Optional, TypeVar

from restate.serde import Serde

I = TypeVar('I')
O = TypeVar('O')

# we will use this symbol to store the handler in the function
RESTATE_UNIQUE_HANDLER_SYMBOL = str(object())

@dataclass
class ServiceTag:
    """
    This class is used to identify the service.
    """
    kind: Literal["object", "service", "workflow"]
    name: str

@dataclass
class HandlerIO(Generic[I, O]):
    """
    Represents the input/output configuration for a handler.

    Attributes:
        accept (str): The accept header value for the handler.
        content_type (str): The content type header value for the handler.
        serializer: The serializer function to convert output to bytes.
        deserializer: The deserializer function to convert input type to bytes.
    """
    accept: str
    content_type: str
    input_serde: Serde[I]
    output_serde: Serde[O]

@dataclass
class Handler(Generic[I, O]):
    """
    Represents a handler for a service.
    """
    service_tag: ServiceTag
    handler_io: HandlerIO[I, O]
    kind: Optional[Literal["exclusive", "shared", "workflow"]]
    name: str
    fn: Callable[[Any, I], Awaitable[O]] | Callable[[Any], Awaitable[O]]
    arity: int


# disable too many arguments warning
# pylint: disable=R0913

def make_handler(service_tag: ServiceTag,
                 handler_io: HandlerIO[I, O],
                 name: str | None,
                 kind: Optional[Literal["exclusive", "shared", "workflow"]],
                 wrapped: Any,
                 arity: int) -> Handler[I, O]:
    """
    Factory function to create a handler.
    """
    # try to deduce the handler name
    handler_name = name
    if not handler_name:
        handler_name = wrapped.__name__
    if not handler_name:
        raise ValueError("Handler name must be provided")

    handler = Handler[I, O](service_tag,
                            handler_io,
                            kind,
                            handler_name,
                            wrapped,
                            arity)
    vars(wrapped)[RESTATE_UNIQUE_HANDLER_SYMBOL] = handler
    return handler

def handler_from_callable(wrapper: Callable[[Any, I], Awaitable[O]]) -> Handler[I, O]:
    """
    Get the handler from the callable.
    """
    try:
        return vars(wrapper)[RESTATE_UNIQUE_HANDLER_SYMBOL]
    except KeyError:
        raise ValueError("Handler not found") # pylint: disable=raise-missing-from

async def invoke_handler(handler: Handler[I, O], ctx: Any, in_buffer: bytes) -> bytes:
    """
    Invoke the handler with the given context and input.
    """
    if handler.arity == 2:
        in_arg = handler.handler_io.input_serde.deserialize(in_buffer) # type: ignore
        out_arg = await handler.fn(ctx, in_arg) # type: ignore [call-arg, arg-type]
    else:
        out_arg = await handler.fn(ctx) # type: ignore [call-arg]
    out_buffer = handler.handler_io.output_serde.serialize(out_arg) # type: ignore
    return bytes(out_buffer)



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""example.py"""
# pylint: disable=C0116
# pylint: disable=W0613
# pylint: disable=W0622

from restate import VirtualObject, ObjectContext
from restate.exceptions import TerminalError

failing = VirtualObject("Failing")

@failing.handler(name="terminallyFailingCall")
async def terminally_failing_call(ctx: ObjectContext, msg: str):
    raise TerminalError(message=msg)

@failing.handler(name="callTerminallyFailingCall")
async def call_terminally_failing_call(ctx: ObjectContext, msg: str) -> str:
    await ctx.object_call(terminally_failing_call,  key="random-583e1bf2", arg=msg)

    raise Exception("Should not reach here")

failures = 0

@failing.handler(name="failingCallWithEventualSuccess")
async def failing_call_with_eventual_success(ctx: ObjectContext) -> int:
    global failures
    failures += 1
    if failures >= 4:
        failures = 0
        return 4
    raise ValueError(f"Failed at attempt: {failures}")


side_effect_failures = 0

@failing.handler(name="failingSideEffectWithEventualSuccess")
async def failing_side_effect_with_eventual_success(ctx: ObjectContext) -> int:

    def side_effect():
        global side_effect_failures
        side_effect_failures += 1
        if side_effect_failures >= 4:
            side_effect_failures = 0
            return 4
        raise ValueError(f"Failed at attempt: {side_effect_failures}")

    return await ctx.run("sideEffect", side_effect) # type: ignore


@failing.handler(name="terminallyFailingSideEffect")
async def terminally_failing_side_effect(ctx: ObjectContext):

    def side_effect():
        raise TerminalError(message="failed side effect")

    await ctx.run("sideEffect", side_effect)
    raise ValueError("Should not reach here")



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""Verification test"""

from datetime import timedelta
import json
from typing import TypedDict
import typing
import random


from restate.context import ObjectContext, ObjectSharedContext
from restate.exceptions import TerminalError
from restate.object import VirtualObject
from restate.serde import JsonSerde
import restate

SET_STATE = 1
GET_STATE = 2
CLEAR_STATE = 3
INCREMENT_STATE_COUNTER = 4
INCREMENT_STATE_COUNTER_INDIRECTLY = 5
SLEEP = 6
CALL_SERVICE = 7
CALL_SLOW_SERVICE = 8
INCREMENT_VIA_DELAYED_CALL = 9
SIDE_EFFECT = 10
THROWING_SIDE_EFFECT = 11
SLOW_SIDE_EFFECT = 12
RECOVER_TERMINAL_CALL = 13
RECOVER_TERMINAL_MAYBE_UN_AWAITED = 14
AWAIT_PROMISE = 15
RESOLVE_AWAKEABLE = 16
REJECT_AWAKEABLE = 17
INCREMENT_STATE_COUNTER_VIA_AWAKEABLE = 18
CALL_NEXT_LAYER_OBJECT = 19

# suppress missing docstring
# pylint: disable=C0115
# pylint: disable=C0116
# pylint: disable=C0301
# pylint: disable=R0914, R0912, R0915, R0913

class SupportService:

    def __init__(self, ctx: ObjectContext) -> None:
        self.ctx = ctx
        self.serde = JsonSerde[typing.Any]()

    async def call(self, method: str, arg: typing.Any) -> typing.Any:
        buffer = self.serde.serialize(arg)
        out_buffer = await self.ctx.generic_call("ServiceInterpreterHelper", method, buffer)
        return self.serde.deserialize(out_buffer)

    def send(self, method: str, arg: typing.Any, delay: int | None = None) -> None:
        buffer = self.serde.serialize(arg)
        if delay is None:
            send_delay = None
        else:
            send_delay = timedelta(milliseconds=delay)
        self.ctx.generic_send("ServiceInterpreterHelper", method, buffer, send_delay=send_delay)

    async def ping(self) -> None:
        return await self.call(method="ping", arg=None)

    async def echo(self, parameters: str) -> str:
        return await self.call(method="echo", arg=parameters)

    async def echo_later(self, parameter: str, sleep: int) -> str:
        arg = {"parameter": parameter, "sleep": sleep}
        return await self.call(method="echo_later", arg=arg)

    async def terminal_failure(self) -> str:
        return await self.call(method="terminalFailure", arg=None)

    async def increment_indirectly(self, layer: int, key: str, delay: typing.Optional[int] = None) -> None:
        arg = {"layer": layer, "key": key}
        self.send(method="incrementIndirectly", arg=arg, delay=delay)

    def resolve_awakeable(self, aid: str) -> None:
        self.send("resolveAwakeable", aid)

    def reject_awakeable(self, aid: str) -> None:
        self.send("rejectAwakeable", aid)

    def increment_via_awakeable_dance(self, layer: int, key: str, tx_promise_id: str) -> None:
        arg = { "interpreter" : { "layer": layer, "key": key} , "txPromiseId": tx_promise_id }
        self.send("incrementViaAwakeableDance", arg)


class Command(TypedDict):
    kind: int
    key: int
    duration: int
    sleep: int
    index: int
    program: typing.Any # avoid circular type


Program = dict[typing.Literal['commands'],
               typing.List[Command]]


async def interpreter(layer: int,
                      ctx: ObjectContext,
                      program: Program) -> None:
    """Interprets a command and executes it."""
    service = SupportService(ctx)
    coros: dict[int,
                typing.Tuple[typing.Any, typing.Awaitable[typing.Any]]] = {}
    for i, command in enumerate(program['commands']):
        print(f"{ctx.request().id} {ctx.key()} {id(ctx)} COMMAND {i} {command}", flush=True)
        command_type = command['kind']
        if command_type == SET_STATE:
            ctx.set(f"key-{command['key']}", f"value-{command['key']}")
        elif command_type == GET_STATE:
            await ctx.get(f"key-{command['key']}")
        elif command_type == CLEAR_STATE:
            ctx.clear(f"key-{command['key']}")
        elif command_type == INCREMENT_STATE_COUNTER:
            c = await ctx.get("counter") or 0
            c += 1
            ctx.set("counter", c)
        elif command_type == SLEEP:
            duration = timedelta(milliseconds=command['duration'])
            await ctx.sleep(duration)
        elif command_type == CALL_SERVICE:
            expected = f"hello-{i}"
            coros[i] = (expected, service.echo(expected))
        elif command_type == INCREMENT_VIA_DELAYED_CALL:
            delay = command['duration']
            await service.increment_indirectly(layer=layer, key=ctx.key(), delay=delay)
        elif command_type == CALL_SLOW_SERVICE:
            expected = f"hello-{i}"
            coros[i] = (expected, service.echo_later(expected, command['sleep']))
        elif command_type == SIDE_EFFECT:
            expected = f"hello-{i}"
            result = await ctx.run("sideEffect",  lambda : expected) # pylint: disable=W0640
            if result != expected:
                raise TerminalError(f"Expected {expected} but got {result}")
        elif command_type == SLOW_SIDE_EFFECT:
            pass
        elif command_type == RECOVER_TERMINAL_CALL:
            try:
                await service.terminal_failure()
            except TerminalError:
                pass
            else:
                raise TerminalError("Expected terminal error")
        elif command_type == RECOVER_TERMINAL_MAYBE_UN_AWAITED:
            pass
        elif command_type == THROWING_SIDE_EFFECT:
            async def side_effect():
                if bool(random.getrandbits(1)):
                    raise ValueError("Random error")

            await ctx.run("throwingSideEffect", side_effect)
        elif command_type == INCREMENT_STATE_COUNTER_INDIRECTLY:
            await service.increment_indirectly(layer=layer, key=ctx.key())
        elif command_type == AWAIT_PROMISE:
            index = command['index']
            if index not in coros:
                continue

            expected, coro = coros[index]
            del coros[index]
            print(f"AWAITING {index} {coro}", flush=True)
            result = await coro
            print (f"AWAITED {index} {result}", flush=True)
            print(f"EXPECTED {result} {expected}", flush=True)
            if result != expected:
                raise TerminalError(f"Expected {expected} but got {result}")
        elif command_type == RESOLVE_AWAKEABLE:
            name, promise = ctx.awakeable()
            coros[i] = ("ok", promise)
            service.resolve_awakeable(name)
        elif command_type == REJECT_AWAKEABLE:
            name, promise = ctx.awakeable()
            coros[i] = ("rejected", promise)
            service.reject_awakeable(name)
        elif command_type == INCREMENT_STATE_COUNTER_VIA_AWAKEABLE:
            tx_promise_id, tx_promise = ctx.awakeable()
            service.increment_via_awakeable_dance(layer=layer, key=ctx.key(), tx_promise_id=tx_promise_id)
            their_promise_for_us_to_resolve: str = await tx_promise
            ctx.resolve_awakeable(their_promise_for_us_to_resolve, "ok")
        elif command_type == CALL_NEXT_LAYER_OBJECT:
            next_layer = f"ObjectInterpreterL{layer + 1}"
            key = f"{command['key']}"
            program = command['program']
            js_program = json.dumps(program)
            raw_js_program = js_program.encode('utf-8')
            promise = ctx.generic_call(next_layer, "interpret", raw_js_program, key)
            print(f"Storing coro {i} {promise}", flush=True)
            coros[i] = (b'', promise)
        else:
            raise ValueError(f"Unknown command type: {command_type}")
        print(f"{ctx.request().id} {ctx.key()} {id(ctx)} DONE command {i}", flush=True)
    print("DONE " + str(len(program['commands'])) , flush=True)

layer_0 = VirtualObject("ObjectInterpreterL0")

@layer_0.handler()
async def interpret(ctx: ObjectContext, program: Program) -> None:
    await interpreter(0, ctx, program)

@layer_0.handler(kind="shared")
async def counter(ctx: ObjectSharedContext) -> int:
    return await ctx.get("counter") or 0


app = restate.app(services=[layer_0])



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""example.py"""
# pylint: disable=C0116
# pylint: disable=W0613

from restate import Service, Context
from typing import TypedDict, Optional, Iterable

proxy = Service("Proxy")


class ProxyRequest(TypedDict):
    serviceName: str
    virtualObjectKey: Optional[str]
    handlerName: str
    message: Iterable[int]


@proxy.handler()
async def call(ctx: Context, req: ProxyRequest) -> Iterable[int]:
    return list(await ctx.generic_call(
        req['serviceName'],
        req['handlerName'],
        bytes(req['message']),
        req['virtualObjectKey']))


@proxy.handler(name="oneWayCall")
async def one_way_call(ctx: Context, req: ProxyRequest):
    ctx.generic_send(
        req['serviceName'],
        req['handlerName'],
        bytes(req['message']),
        req['virtualObjectKey'])


class ManyCallRequest(TypedDict):
    proxyRequest: ProxyRequest
    oneWayCall: bool
    awaitAtTheEnd: bool

@proxy.handler(name="manyCalls")
async def many_calls(ctx: Context, requests: Iterable[ManyCallRequest]):
    to_await = []

    for req in requests:
        if req['oneWayCall']:
            ctx.generic_send(
                req['proxyRequest']['serviceName'],
                req['proxyRequest']['handlerName'],
                bytes(req['proxyRequest']['message']),
                req['proxyRequest']['virtualObjectKey'])
        else:
            awaitable = ctx.generic_call(
                req['proxyRequest']['serviceName'],
                req['proxyRequest']['handlerName'],
                bytes(req['proxyRequest']['message']),
                req['proxyRequest']['virtualObjectKey'])
            if req['awaitAtTheEnd']:
                to_await.append(awaitable)

    for awaitable in to_await:
        await awaitable



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""This module contains the restate context implementation based on the server"""

from datetime import timedelta
import inspect
from typing import Any, Awaitable, Callable, Dict, List, Optional, TypeVar
import typing
import traceback

from restate.context import DurablePromise, ObjectContext, Request
from restate.exceptions import TerminalError
from restate.handler import Handler, handler_from_callable, invoke_handler
from restate.serde import BytesSerde, JsonSerde, Serde
from restate.server_types import Receive, Send
from restate.vm import Failure, Invocation, NotReady, SuspendedException, VMWrapper


T = TypeVar('T')
I = TypeVar('I')
O = TypeVar('O')

# disable to many arguments
# pylint: disable=R0913

# disable line too long
# pylint: disable=C0301

async def async_value(n: Callable[[], T]) -> T:
    """convert a simple value to a coroutine."""
    return n()


class ServerDurablePromise(DurablePromise):
    """This class implements a durable promise API"""

    def __init__(self, server_context, name, serde) -> None:
        super().__init__(name=name, serde=JsonSerde() if serde is None else serde)
        self.server_context = server_context

    def value(self) -> Awaitable[Any]:
        vm: VMWrapper = self.server_context.vm
        handle = vm.sys_get_promise(self.name)
        coro =  self.server_context.create_poll_coroutine(handle)
        serde = self.serde
        assert serde is not None

        async def await_point():
            res = await coro
            return serde.deserialize(res)

        return await_point()

    def resolve(self, value: Any) -> Awaitable[None]:
        vm: VMWrapper = self.server_context.vm
        assert self.serde is not None
        value_buffer = self.serde.serialize(value)
        handle = vm.sys_complete_promise_success(self.name, value_buffer)
        return self.server_context.create_poll_coroutine(handle)

    def reject(self, message: str, code: int = 500) -> Awaitable[None]:
        vm: VMWrapper = self.server_context.vm
        py_failure = Failure(code=code, message=message)
        handle = vm.sys_complete_promise_failure(self.name, py_failure)
        return self.server_context.create_poll_coroutine(handle)

    def peek(self) -> Awaitable[Any | None]:
        vm: VMWrapper = self.server_context.vm
        handle = vm.sys_peek_promise(self.name)
        coro =  self.server_context.create_poll_coroutine(handle)
        serde = self.serde
        assert serde is not None

        async def await_point():
            res = await coro
            if res is None:
                return None
            return serde.deserialize(res)

        return await_point()


# disable too many public method
# pylint: disable=R0904

class ServerInvocationContext(ObjectContext):
    """This class implements the context for the restate framework based on the server."""

    def __init__(self,
                 vm: VMWrapper,
                 handler: Handler[I, O],
                 invocation: Invocation,
                 attempt_headers: Dict[str, str],
                 send: Send,
                 receive: Receive) -> None:
        super().__init__()
        self.vm = vm
        self.handler = handler
        self.invocation = invocation
        self.attempt_headers = attempt_headers
        self.send = send
        self.receive = receive

    async def enter(self):
        """Invoke the user code."""
        try:
            in_buffer = self.invocation.input_buffer
            out_buffer = await invoke_handler(handler=self.handler, ctx=self, in_buffer=in_buffer)
            self.vm.sys_write_output_success(bytes(out_buffer))
            self.vm.sys_end()
        except TerminalError as t:
            failure = Failure(code=t.status_code, message=t.message)
            self.vm.sys_write_output_failure(failure)
            self.vm.sys_end()
        # pylint: disable=W0718
        except SuspendedException:
            pass
        except Exception as e:
            fmt = '\n'.join(traceback.format_exception(e))
            self.vm.notify_error(fmt)
            raise e

    async def leave(self):
        """Leave the context."""
        while True:
            chunk = self.vm.take_output()
            if chunk is None:
                break
            await self.send({
                'type': 'http.response.body',
                'body': chunk,
                'more_body': True,
            })
        # ========================================
        # End the connection
        # ========================================
        # The following sequence of events is expected during a teardown:
        #
        # {'type': 'http.request', 'body': b'', 'more_body': True}
        # {'type': 'http.request', 'body': b'', 'more_body': False}
        # {'type': 'http.disconnect'}
        while True:
            event = await self.receive()
            if event is None:
                break
            if event.get('type') == 'http.disconnect':
                break
            if event.get('type') == 'http.request' and event.get('more_body', False) is False:
                break
        # finally, we close our side
        # it is important to do it, after the other side has closed his side,
        # because some asgi servers (like hypercorn) will remove the stream
        # as soon as they see a close event (in asgi terms more_body=False)
        await self.send({
            'type': 'http.response.body',
            'body': b'',
            'more_body': False,
        })


    async def create_poll_coroutine(self, handle) -> bytes | None:
        """Create a coroutine to poll the handle."""
        output = self.vm.take_output()
        if output:
            await self.send({
                'type': 'http.response.body',
                'body': bytes(output),
                'more_body': True,
            })
        self.vm.notify_await_point(handle)
        while True:
            res = self.vm.take_async_result(handle)
            if isinstance(res, NotReady):
                chunk = await self.receive()
                if chunk.get('body', None) is not None:
                    assert isinstance(chunk['body'], bytes)
                    self.vm.notify_input(chunk['body'])
                if not chunk.get('more_body', False):
                    self.vm.notify_input_closed()
                continue
            if res is None:
                return None
            if isinstance(res, Failure):
                raise TerminalError(res.message, res.code)
            return res

    def get(self, name: str, serde: Serde[T] = JsonSerde()) -> typing.Awaitable[Optional[Any]]:
        coro = self.create_poll_coroutine(self.vm.sys_get_state(name))

        async def await_point():
            """Wait for this handle to be resolved."""
            res = await coro
            if res is None:
                return None
            return serde.deserialize(res)

        return await_point() # do not await here, the caller will do it.

    def state_keys(self) -> Awaitable[List[str]]:
        return self.create_poll_coroutine(self.vm.sys_get_state_keys()) # type: ignore

    def set(self, name: str, value: T, serde: Serde[T] = JsonSerde()) -> None:
        """Set the value associated with the given name."""
        buffer = serde.serialize(value)
        self.vm.sys_set_state(name, bytes(buffer))

    def clear(self, name: str) -> None:
        self.vm.sys_clear_state(name)

    def clear_all(self) -> None:
        self.vm.sys_clear_all_state()

    def request(self) -> Request:
        return Request(
            id=self.invocation.invocation_id,
            headers=dict(self.invocation.headers),
            attempt_headers=self.attempt_headers,
            body=self.invocation.input_buffer,
        )

    # pylint: disable=W0236
    async def run(self,
                  name: str,
                  action: Callable[[], T] | Callable[[], Awaitable[T]],
                  serde: Optional[Serde[T]] = JsonSerde()) -> T | None:
        assert serde is not None
        res = self.vm.sys_run_enter(name)
        if isinstance(res, Failure):
            raise TerminalError(res.message, res.code)
        if isinstance(res, bytes):
            return serde.deserialize(res)
        # the side effect was not executed before, so we need to execute it now
        assert res is None
        try:
            if inspect.iscoroutinefunction(action):
                action_result = await action() # type: ignore
            else:
                action_result = action()
            buffer = serde.serialize(action_result)
            handle = self.vm.sys_run_exit_success(buffer)
            await self.create_poll_coroutine(handle)
            return action_result
        except TerminalError as t:
            failure = Failure(code=t.status_code, message=t.message)
            handle = self.vm.sys_run_exit_failure(failure)
            await self.create_poll_coroutine(handle)
            # unreachable
            assert False

    def sleep(self, delta: timedelta) -> Awaitable[None]:
        # convert timedelta to milliseconds
        millis = int(delta.total_seconds() * 1000)
        return self.create_poll_coroutine(self.vm.sys_sleep(millis)) # type: ignore

    def do_call(self,
                tpe: Callable[[Any, I], Awaitable[O]],
                parameter: I,
                key: Optional[str] = None,
                send_delay: Optional[timedelta] = None,
                send: bool = False) -> Awaitable[O] | None:
        """Make an RPC call to the given handler"""
        target_handler = handler_from_callable(tpe)
        service=target_handler.service_tag.name
        handler=target_handler.name
        input_serde = target_handler.handler_io.input_serde
        output_serde = target_handler.handler_io.output_serde
        return self.do_raw_call(service, handler, parameter, input_serde, output_serde, key, send_delay, send)


    def do_raw_call(self,
                 service: str,
                 handler:str,
                 input_param: I,
                 input_serde: Serde[I],
                 output_serde: Serde[O],
                 key: Optional[str] = None,
                 send_delay: Optional[timedelta] = None,
                 send: bool = False) -> Awaitable[O] | None:
        """Make an RPC call to the given handler"""
        parameter = input_serde.serialize(input_param)
        if send_delay:
            ms = int(send_delay.total_seconds() * 1000)
            self.vm.sys_send(service, handler, parameter, key, delay=ms)
            return None
        if send:
            self.vm.sys_send(service, handler, parameter, key)
            return None

        handle = self.vm.sys_call(service=service,
                                  handler=handler,
                                  parameter=parameter,
                                  key=key)

        async def await_point(s: ServerInvocationContext, h, o: Serde[O]):
            """Wait for this handle to be resolved, and deserialize the response."""
            res = await s.create_poll_coroutine(h)
            return o.deserialize(res) # type: ignore

        return await_point(self, handle, output_serde)

    def service_call(self,
                     tpe: Callable[[Any, I], Awaitable[O]],
                     arg: I) -> Awaitable[O]:
        coro = self.do_call(tpe, arg)
        assert coro is not None
        return coro

    def service_send(self, tpe: Callable[[Any, I], Awaitable[O]], arg: I, send_delay: timedelta | None = None) -> None:
        self.do_call(tpe=tpe, parameter=arg, send_delay=send_delay, send=True)

    def object_call(self,
                    tpe: Callable[[Any, I],Awaitable[O]],
                    key: str,
                    arg: I,
                    send_delay: Optional[timedelta] = None,
                    send: bool = False) -> Awaitable[O]:
        coro = self.do_call(tpe, arg, key, send_delay, send)
        assert coro is not None
        return coro

    def object_send(self, tpe: Callable[[Any, I], Awaitable[O]], key: str, arg: I, send_delay: timedelta | None = None) -> None:
        self.do_call(tpe=tpe, key=key, parameter=arg, send_delay=send_delay, send=True)

    def workflow_call(self,
                        tpe: Callable[[Any, I], Awaitable[O]],
                        key: str,
                        arg: I) -> Awaitable[O]:
        return self.object_call(tpe, key, arg)

    def workflow_send(self, tpe: Callable[[Any, I], Awaitable[O]], key: str, arg: I, send_delay: timedelta | None = None) -> None:
        return self.object_send(tpe, key, arg, send_delay)

    def generic_call(self, service: str, handler: str, arg: bytes, key: str | None = None) -> Awaitable[bytes]:
        serde = BytesSerde()
        return self.do_raw_call(service, handler, arg, serde, serde, key) # type: ignore

    def generic_send(self, service: str, handler: str, arg: bytes, key: str | None = None, send_delay: timedelta | None = None) -> None:
        serde = BytesSerde()
        return self.do_raw_call(service, handler, arg, serde, serde , key, send_delay, True) # type: ignore

    def awakeable(self,
                  serde: typing.Optional[Serde[I]] = JsonSerde()) -> typing.Tuple[str, Awaitable[Any]]:
        assert serde is not None
        name, handle = self.vm.sys_awakeable()
        coro = self.create_poll_coroutine(handle)

        async def await_point():
            """Wait for this handle to be resolved."""
            res = await coro
            assert res is not None
            return serde.deserialize(res)


        return name, await_point()

    def resolve_awakeable(self,
                          name: str,
                          value: I,
                          serde: typing.Optional[Serde[I]] = JsonSerde()) -> None:
        assert serde is not None
        buf = serde.serialize(value)
        self.vm.sys_resolve_awakeable(name, buf)

    def reject_awakeable(self, name: str, failure_message: str, failure_code: int = 500) -> None:
        return self.vm.sys_reject_awakeable(name, Failure(code=failure_code, message=failure_message))

    def promise(self, name: str, serde: typing.Optional[Serde[T]] = JsonSerde()) -> DurablePromise[Any]:
        """Create a durable promise."""
        return ServerDurablePromise(self, name, serde)

    def key(self) -> str:
        return self.invocation.key



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""example.py"""
# pylint: disable=C0116
# pylint: disable=W0613

from typing import TypedDict
from restate import VirtualObject, ObjectContext

map_object = VirtualObject("MapObject")


class Entry(TypedDict):
    key: str
    value: str

@map_object.handler(name="set")
async def map_set(ctx: ObjectContext, entry: Entry):
    ctx.set(entry["key"], entry["value"])

@map_object.handler(name="get")
async def map_get(ctx: ObjectContext, key: str) -> str:
    return await ctx.get(key) or ""

@map_object.handler(name="clearAll")
async def map_clear_all(ctx: ObjectContext) -> list[Entry]:
    entries = []
    for key in await ctx.state_keys():
        value: str = await ctx.get(key) # type: ignore 
        entry = Entry(key=key, value=value)
        entries.append(entry)
        ctx.clear(key)
    return entries


#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""example.py"""
# pylint: disable=C0116
# pylint: disable=W0613

from restate import VirtualObject, ObjectContext

list_object = VirtualObject("ListObject")

@list_object.handler()
async def append(ctx: ObjectContext, value: str):
    list = await ctx.get("list") or []
    ctx.set("list", list + [value])

@list_object.handler()
async def get(ctx: ObjectContext) -> list[str]:
    return await ctx.get("list") or []

@list_object.handler()
async def clear(ctx: ObjectContext) -> list[str]:
    result = await ctx.get("list") or []
    ctx.clear("list")
    return result



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""example.py"""
# pylint: disable=C0116
# pylint: disable=W0613
# pylint: disable=C0301


from restate import Workflow, WorkflowContext, WorkflowSharedContext
from restate.exceptions import TerminalError

payment = Workflow("payment")

@payment.main()
async def pay(ctx: WorkflowContext, amount: int):
    workflow_key = ctx.key()
    ctx.set("status", "verifying payment")

    # Call the payment service
    def payment_gateway():
        print("Please approve this payment: ")
        print("To approve use:")
        print(f"""curl http://localhost:8080/payment/{workflow_key}/payment_verified --json '"approved"' """)
        print("")
        print("To decline use:")
        print(f"""curl http://localhost:8080/payment/{workflow_key}/payment_verified --json '"declined"' """)

    await ctx.run("payment", payment_gateway)

    ctx.set("status", "waiting for the payment provider to approve")

    # Wait for the payment to be verified
    result = await ctx.promise("verify.payment").value()
    if result == "approved":
        ctx.set("status", "payment approved")
        return { "success" : True }

    ctx.set("status", "payment declined")
    raise TerminalError(message="Payment declined", status_code=401)

@payment.handler()
async def payment_verified(ctx: WorkflowSharedContext, result: str):
    promise = ctx.promise("verify.payment")
    await promise.resolve(result)



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
This module contains the Lambda/ASGI adapter.
"""
import asyncio
import base64
import os
from typing import TypedDict, Dict, cast, Union, Any, Callable

from restate.server_types import (ASGIApp,
                                  Scope,
                                  Receive,
                                  HTTPResponseStartEvent,
                                  HTTPResponseBodyEvent,
                                  HTTPRequestEvent)

class RestateLambdaRequest(TypedDict):
    """
    Restate Lambda request

    :see: https://github.com/restatedev/restate/blob/1a10c05b16b387191060b49faffb0335ee97e96d/crates/service-client/src/lambda.rs#L297 # pylint: disable=line-too-long
    """
    path: str
    httpMethod: str
    headers: Dict[str, str]
    body: str
    isBase64Encoded: bool


class RestateLambdaResponse(TypedDict):
    """
    Restate Lambda response

    :see: https://github.com/restatedev/restate/blob/1a10c05b16b387191060b49faffb0335ee97e96d/crates/service-client/src/lambda.rs#L310 # pylint: disable=line-too-long
    """
    statusCode: int
    headers: Dict[str, str]
    body: str
    isBase64Encoded: bool


RestateLambdaHandler = Callable[[RestateLambdaRequest, Any], RestateLambdaResponse]


def create_scope(req: RestateLambdaRequest) -> Scope:
    """
    Create ASGI scope from lambda request
    """
    headers = {k.lower(): v for k, v in req.get('headers', {}).items()}
    http_method = req["httpMethod"]
    path = req["path"]

    return {
        "type": "http",
        "method": http_method,
        "http_version": "1.1",
        "headers": [(k.encode(), v.encode()) for k, v in headers.items()],
        "path": path,
        "scheme": headers.get("x-forwarded-proto", "https"),
        "asgi": {"version": "3.0", "spec_version": "2.0"},
        "raw_path": path.encode(),
        "root_path": "",
        "query_string": b'',
        "client": None,
        "server": None,
        "extensions": None
    }


def request_to_receive(req: RestateLambdaRequest) -> Receive:
    """
    Create ASGI Receive from lambda request
    """
    assert req['isBase64Encoded']
    body = base64.b64decode(req['body'])

    events = cast(list[HTTPRequestEvent], [{
        "type": "http.request",
        "body": body,
        "more_body": False
    }])

    async def recv() -> HTTPRequestEvent:
        if len(events) != 0:
            return events.pop()
        return {
            "type": "http.request",
            "body": b'',
            "more_body": False
        }

    return recv


class ResponseCollector:
    """
    Response collector from ASGI Send to Lambda
    """
    def __init__(self):
        self.body = bytearray()
        self.headers = {}
        self.status_code = 500

    async def __call__(self, message: Union[HTTPResponseStartEvent, HTTPResponseBodyEvent]) -> None:
        """
        Implements ASGI send contract
        """
        if message["type"] == "http.response.start":
            self.status_code = cast(int, message["status"])
            self.headers = {
                key.decode("utf-8"): value.decode("utf-8")
                for key, value in message["headers"]
            }
        elif message["type"] == "http.response.body" and "body" in message:
            self.body.extend(message["body"])
        return

    def to_lambda_response(self) -> RestateLambdaResponse:
        """
        Convert collected values to lambda response
        """
        return {
            "statusCode": self.status_code,
            "headers": self.headers,
            "isBase64Encoded": True,
            "body": base64.b64encode(self.body).decode()
        }


def is_running_on_lambda() -> bool:
    """
    :return: true if this Python script is running on Lambda
    """
    # https://docs.aws.amazon.com/lambda/latest/dg/configuration-envvars.html
    return "AWS_LAMBDA_FUNCTION_NAME" in os.environ


def wrap_asgi_as_lambda_handler(asgi_app: ASGIApp) \
        -> Callable[[RestateLambdaRequest, Any], RestateLambdaResponse]:
    """
    Wrap the given asgi_app in a Lambda handler
    """
    # Setup AsyncIO
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    def lambda_handler(event: RestateLambdaRequest, _context: Any) -> RestateLambdaResponse:
        loop = asyncio.get_event_loop()

        scope = create_scope(event)
        recv = request_to_receive(event)
        send = ResponseCollector()

        asgi_instance = asgi_app(scope, recv, send)
        asgi_task = loop.create_task(asgi_instance)  # type: ignore[var-annotated, arg-type]
        loop.run_until_complete(asgi_task)

        return send.to_lambda_response()

    return lambda_handler



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""example.py"""
# pylint: disable=C0116
# pylint: disable=W0613

from datetime import timedelta
from typing import Literal

from restate import VirtualObject, ObjectContext
from restate.exceptions import TerminalError

from . import awakable_holder

BlockingOperation = Literal["CALL", "SLEEP", "AWAKEABLE"]

runner = VirtualObject("CancelTestRunner")

@runner.handler(name="startTest")
async def start_test(ctx: ObjectContext, op: BlockingOperation):
    try:
        await ctx.object_call(block, key="", arg=op)
    except TerminalError as t:
        if t.status_code == 409:
            ctx.set("state", True)
        else:
            raise t

@runner.handler(name="verifyTest")
async def verify_test(ctx: ObjectContext) -> bool:
    state = await ctx.get("state")
    if state is None:
        return False
    return state

                                  
blocking_service = VirtualObject("CancelTestBlockingService")

@blocking_service.handler()
async def block(ctx: ObjectContext, op: BlockingOperation):
    name, awakeable = ctx.awakeable()
    await ctx.object_call(awakable_holder.hold, key="cancel", arg=name)
    await awakeable

    if op == "CALL":
        await ctx.object_call(block, key="", arg=op)
    elif op == "SLEEP":
        await ctx.sleep(timedelta(days=1024))
    elif op == "AWAKEABLE":
        name, uncompleteable = ctx.awakeable()
        await uncompleteable

@blocking_service.handler(name="isUnlocked")
async def is_unlocked(ctx: ObjectContext):
    return None



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
Restate Context
"""

import abc
from dataclasses import dataclass
from typing import Any, Awaitable, Callable, Dict, List, Optional, TypeVar, Union
import typing
from datetime import timedelta
from restate.serde import JsonSerde, Serde

T = TypeVar('T')
I = TypeVar('I')
O = TypeVar('O')

RunAction = Union[Callable[[], T], Callable[[], Awaitable[T]]]

@dataclass
class Request:
    """
    Represents an ingress request.

    Attributes:
        id: The unique identifier of the request.
        headers: The headers of the request.
        attempt_headers: The attempt headers of the request.
        body (bytes): The body of the request.
    """
    id: str
    headers: Dict[str, str]
    attempt_headers: Dict[str,str]
    body: bytes


class KeyValueStore(abc.ABC):
    """
    A key scoped key-value store.

    This class defines the interface for a key-value store,
    which allows storing and retrieving values
    based on a unique key.

    """

    @abc.abstractmethod
    def get(self,
            name: str,
            serde: Serde[T] = JsonSerde()) -> Awaitable[Optional[Any]]:
        """
        Retrieves the value associated with the given name.
        """

    @abc.abstractmethod
    def state_keys(self) -> Awaitable[List[str]]:
        """Returns the list of keys in the store."""

    @abc.abstractmethod
    def set(self,
            name: str,
            value: T,
            serde: Serde[T] = JsonSerde()) -> None:
        """set the value associated with the given name."""

    @abc.abstractmethod
    def clear(self, name: str) -> None:
        """clear the value associated with the given name."""

    @abc.abstractmethod
    def clear_all(self) -> None:
        """clear all the values in the store."""


class Context(abc.ABC):
    """
    Represents the context of the current invocation.
    """

    @abc.abstractmethod
    def request(self) -> Request:
        """
        Returns the request object.
        """

    @abc.abstractmethod
    def run(self,
            name: str,
            action: RunAction[T],
            serde: Serde[T] = JsonSerde()) -> Awaitable[T | None]:
        """
        Runs the given action with the given name.
        """

    @abc.abstractmethod
    def sleep(self, delta: timedelta) -> Awaitable[None]:
        """
        Suspends the current invocation for the given duration
        """

    @abc.abstractmethod
    def service_call(self,
                     tpe: Callable[[Any, I], Awaitable[O]],
                     arg: I) -> Awaitable[O]:
        """
        Invokes the given service with the given argument.
        """


    @abc.abstractmethod
    def service_send(self,
                     tpe: Callable[[Any, I], Awaitable[O]],
                     arg: I,
                     send_delay: Optional[timedelta] = None,
                     ) -> None:
        """
        Invokes the given service with the given argument.
        """

    @abc.abstractmethod
    def object_call(self,
                    tpe: Callable[[Any, I], Awaitable[O]],
                    key: str,
                    arg: I) -> Awaitable[O]:
        """
        Invokes the given object with the given argument.
        """

    @abc.abstractmethod
    def object_send(self,
                    tpe: Callable[[Any, I], Awaitable[O]],
                    key: str,
                    arg: I,
                    send_delay: Optional[timedelta] = None,
                    ) -> None:
        """
        Send a message to an object with the given argument.
        """

    @abc.abstractmethod
    def workflow_call(self,
                    tpe: Callable[[Any, I], Awaitable[O]],
                    key: str,
                    arg: I) -> Awaitable[O]:
        """
        Invokes the given workflow with the given argument.
        """

    @abc.abstractmethod
    def workflow_send(self,
                    tpe: Callable[[Any, I], Awaitable[O]],
                    key: str,
                    arg: I,
                    send_delay: Optional[timedelta] = None,
                    ) -> None:
        """
        Send a message to an object with the given argument.
        """

    # pylint: disable=R0913
    @abc.abstractmethod
    def generic_call(self,
                     service: str,
                     handler: str,
                     arg: bytes,
                     key: Optional[str] = None)  -> Awaitable[bytes]:
        """
        Invokes the given generic service/handler with the given argument.
        """

    @abc.abstractmethod
    def generic_send(self,
                     service: str,
                     handler: str,
                     arg: bytes,
                     key: Optional[str] = None,
                     send_delay: Optional[timedelta] = None) -> None:
        """
        Send a message to a generic service/handler with the given argument.
        """

    @abc.abstractmethod
    def awakeable(self,
                  serde: Serde[T] = JsonSerde()) -> typing.Tuple[str, Awaitable[Any]]:
        """
        Returns the name of the awakeable and the future to be awaited.
        """

    @abc.abstractmethod
    def resolve_awakeable(self,
                          name: str,
                          value: I,
                          serde: Serde[I] = JsonSerde()) -> None:
        """
        Resolves the awakeable with the given name.
        """

    @abc.abstractmethod
    def reject_awakeable(self, name: str, failure_message: str, failure_code: int = 500) -> None:
        """
        Rejects the awakeable with the given name.
        """


class ObjectContext(Context, KeyValueStore):
    """
    Represents the context of the current invocation.
    """

    @abc.abstractmethod
    def key(self) -> str:
        """
        Returns the key of the current object.
        """


class ObjectSharedContext(Context):
    """
    Represents the context of the current invocation.
    """

    @abc.abstractmethod
    def key(self) -> str:
        """Returns the key of the current object."""

    @abc.abstractmethod
    def get(self,
            name: str,
            serde: Serde[T] = JsonSerde()) -> Awaitable[Optional[Any]]:
        """
        Retrieves the value associated with the given name.
        """

    @abc.abstractmethod
    def state_keys(self) -> Awaitable[List[str]]:
        """
        Returns the list of keys in the store.
        """

class DurablePromise(typing.Generic[T]):
    """
    Represents a durable promise.
    """

    def __init__(self, name: str, serde: Serde[T] = JsonSerde()) -> None:
        self.name = name
        self.serde = serde

    @abc.abstractmethod
    def resolve(self, value: T) -> Awaitable[None]:
        """
        Resolves the promise with the given value.
        """

    @abc.abstractmethod
    def reject(self, message: str, code: int = 500) -> Awaitable[None]:
        """
        Rejects the promise with the given message and code.
        """

    @abc.abstractmethod
    def peek(self) -> Awaitable[typing.Optional[T]]:
        """
        Returns the value of the promise if it is resolved, None otherwise.
        """

    @abc.abstractmethod
    def value(self) -> Awaitable[T]:
        """
        Returns the value of the promise if it is resolved, None otherwise.
        """

class WorkflowContext(ObjectContext):
    """
    Represents the context of the current workflow invocation.
    """

    @abc.abstractmethod
    def promise(self, name: str, serde: Serde[T] = JsonSerde()) -> DurablePromise[Any]:
        """
        Returns a durable promise with the given name.
        """

class WorkflowSharedContext(ObjectSharedContext):
    """
    Represents the context of the current workflow invocation.
    """

    @abc.abstractmethod
    def promise(self, name: str, serde: Serde[T] = JsonSerde()) -> DurablePromise[Any]:
        """
        Returns a durable promise with the given name.
        """



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""example.py"""
# pylint: disable=C0116
# pylint: disable=W0613

from restate import VirtualObject, ObjectContext, ObjectSharedContext

counter = VirtualObject("counter")

@counter.handler()
async def increment(ctx: ObjectContext, value: int) -> int:
    n = await ctx.get("counter") or 0
    n += value
    ctx.set("counter", n)
    return n

@counter.handler(kind="shared")
async def count(ctx: ObjectSharedContext) -> int:
    return await ctx.get("counter") or 0



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""example.py"""
# pylint: disable=C0116
# pylint: disable=W0613

from greeter import greeter
from virtual_object import counter
from workflow import payment

import restate

app = restate.app(services=[greeter, counter, payment])



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""example.py"""
# pylint: disable=C0116
# pylint: disable=W0613
# pylint: disable=W0622

from restate import Workflow, WorkflowContext, WorkflowSharedContext
from restate.exceptions import TerminalError

workflow = Workflow("BlockAndWaitWorkflow")

@workflow.main()
async def run(ctx: WorkflowContext, input: str):
    ctx.set("my-state", input)
    output = await ctx.promise("durable-promise").value()

    peek = await ctx.promise("durable-promise").peek()
    if peek is None:
        raise TerminalError(message="Durable promise should be completed")

    return output


@workflow.handler()
async def unblock(ctx: WorkflowSharedContext, output: str):
    await ctx.promise("durable-promise").resolve(output)

@workflow.handler(name="getState")
async def get_state(ctx: WorkflowSharedContext, output: str) -> str | None:
    return await ctx.get("my-state")



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""example.py"""
# pylint: disable=C0116
# pylint: disable=W0613

from typing import TypedDict
from restate import VirtualObject, ObjectContext
from restate.exceptions import TerminalError

counter_object = VirtualObject("Counter")

COUNTER_KEY = "counter"


@counter_object.handler()
async def reset(ctx: ObjectContext):
    ctx.clear(COUNTER_KEY)


@counter_object.handler()
async def get(ctx: ObjectContext) -> int:
    c: int | None = await ctx.get(COUNTER_KEY)
    if c is None:
        return 0
    return c


class CounterUpdateResponse(TypedDict):
    oldValue: int
    newValue: int


@counter_object.handler()
async def add(ctx: ObjectContext, addend: int) -> CounterUpdateResponse:
    old_value: int | None = await ctx.get(COUNTER_KEY)
    if old_value is None:
        old_value = 0
    new_value = old_value + addend
    ctx.set(COUNTER_KEY, new_value)
    return CounterUpdateResponse(oldValue=old_value, newValue=new_value)


@counter_object.handler(name="addThenFail")
async def add_then_fail(ctx: ObjectContext, addend: int):
    old_value: int | None = await ctx.get(COUNTER_KEY)
    if old_value is None:
        old_value = 0
    new_value = old_value + addend
    ctx.set(COUNTER_KEY, new_value)

    raise TerminalError(message=ctx.key())


#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
This module contains the ASGI types definitions.

:see: https://github.com/django/asgiref/blob/main/asgiref/typing.py
"""

from typing import (Awaitable, Callable, Dict, Iterable, List,
                    Tuple, Union, TypedDict, Literal, Optional, NotRequired, Any)

class ASGIVersions(TypedDict):
    """ASGI Versions"""
    spec_version: str
    version: Union[Literal["2.0"], Literal["3.0"]]

class Scope(TypedDict):
    """ASGI Scope"""
    type: Literal["http"]
    asgi: ASGIVersions
    http_version: str
    method: str
    scheme: str
    path: str
    raw_path: bytes
    query_string: bytes
    root_path: str
    headers: Iterable[Tuple[bytes, bytes]]
    client: Optional[Tuple[str, int]]
    server: Optional[Tuple[str, Optional[int]]]
    state: NotRequired[Dict[str, Any]]
    extensions: Optional[Dict[str, Dict[object, object]]]

class HTTPRequestEvent(TypedDict):
    """ASGI Request event"""
    type: Literal["http.request"]
    body: bytes
    more_body: bool

class HTTPResponseStartEvent(TypedDict):
    """ASGI Response start event"""
    type: Literal["http.response.start"]
    status: int
    headers: Iterable[Tuple[bytes, bytes]]
    trailers: bool

class HTTPResponseBodyEvent(TypedDict):
    """ASGI Response body event"""
    type: Literal["http.response.body"]
    body: bytes
    more_body: bool


ASGIReceiveEvent = Union[
    HTTPRequestEvent
]


ASGISendEvent = Union[
    HTTPResponseStartEvent,
    HTTPResponseBodyEvent
]

Receive = Callable[[], Awaitable[ASGIReceiveEvent]]
Send = Callable[[ASGISendEvent], Awaitable[None]]

ASGIApp = Callable[
    [
        Scope,
        Receive,
        Send,
    ],
    Awaitable[None],
]

def header_to_binary(headers: Iterable[Tuple[str, str]]) -> List[Tuple[bytes, bytes]]:
    """Convert a list of headers to a list of binary headers."""
    return [ (k.encode('utf-8'), v.encode('utf-8')) for k,v in headers ]

def binary_to_header(headers: Iterable[Tuple[bytes, bytes]]) -> List[Tuple[str, str]]:
    """Convert a list of binary headers to a list of headers."""
    return [ (k.decode('utf-8'), v.decode('utf-8')) for k,v in headers ]



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""example.py"""
# pylint: disable=C0116
# pylint: disable=W0613
# pylint: disable=W0622

from restate import VirtualObject, ObjectContext
from restate.exceptions import TerminalError

awakeable_holder = VirtualObject("AwakeableHolder")

@awakeable_holder.handler()
async def hold(ctx: ObjectContext, id: str):
    ctx.set("id", id)

@awakeable_holder.handler(name="hasAwakeable")
async def has_awakeable(ctx: ObjectContext) -> bool:
    res = await ctx.get("id")
    return res is not None

@awakeable_holder.handler()
async def unlock(ctx: ObjectContext, payload: str):
    id = await ctx.get("id")
    if id is None:
        raise TerminalError(message="No awakeable is registered")
    ctx.resolve_awakeable(id, payload)



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
Holds the discovery API objects as defined by the restate protocol.
Note that the classes defined here do not use snake case, because they 
are intended to be serialized to JSON, and their cases must remain in the 
case that the restate server understands.
"""

# disable to few parameters
# pylint: disable=R0903
# pylint: disable=C0301
# pylint: disable=C0115
# pylint: disable=C0103
# pylint: disable=W0622

import json
import typing
from enum import Enum
from typing import Optional, Any, List


from restate.endpoint import Endpoint as RestateEndpoint

class ProtocolMode(Enum):
    BIDI_STREAM = "BIDI_STREAM"
    REQUEST_RESPONSE = "REQUEST_RESPONSE"

class ServiceType(Enum):
    VIRTUAL_OBJECT = "VIRTUAL_OBJECT"
    SERVICE = "SERVICE"
    WORKFLOW = "WORKFLOW"

class ServiceHandlerType(Enum):
    WORKFLOW = "WORKFLOW"
    EXCLUSIVE = "EXCLUSIVE"
    SHARED = "SHARED"

class InputPayload:
    def __init__(self, required: bool, contentType: str, jsonSchema: Optional[Any] = None):
        self.required = required
        self.contentType = contentType
        self.jsonSchema = jsonSchema

class OutputPayload:
    def __init__(self, contentType: str, setContentTypeIfEmpty: bool, jsonSchema: Optional[Any] = None):
        self.contentType = contentType
        self.setContentTypeIfEmpty = setContentTypeIfEmpty
        self.jsonSchema = jsonSchema

class Handler:
    def __init__(self, name: str, ty: Optional[ServiceHandlerType] = None, input: Optional[InputPayload] = None, output: Optional[OutputPayload] = None):
        self.name = name
        self.ty = ty
        self.input = input
        self.output = output

class Service:
    def __init__(self, name: str, ty: ServiceType, handlers: List[Handler]):
        self.name = name
        self.ty = ty
        self.handlers = handlers

class Endpoint:
    def __init__(self, protocolMode: ProtocolMode, minProtocolVersion: int, maxProtocolVersion: int, services: List[Service]):
        self.protocolMode = protocolMode
        self.minProtocolVersion = minProtocolVersion
        self.maxProtocolVersion = maxProtocolVersion
        self.services = services

PROTOCOL_MODES = {
        "bidi" : ProtocolMode.BIDI_STREAM,
        "request_response" : ProtocolMode.REQUEST_RESPONSE}

SERVICE_TYPES = {
            "service": ServiceType.SERVICE,
            "object": ServiceType.VIRTUAL_OBJECT,
            "workflow": ServiceType.WORKFLOW}

HANDLER_TYPES  = {
            'exclusive': ServiceHandlerType.EXCLUSIVE,
            'shared': ServiceHandlerType.SHARED,
            'workflow': ServiceHandlerType.WORKFLOW}

class PythonClassEncoder(json.JSONEncoder):
    """
    Serialize Python objects as JSON
    """
    def default(self, o):
        if isinstance(o, Enum):
            return o.value
        return {key: value for key, value in o.__dict__.items() if value is not None}

def compute_discovery_json(endpoint: RestateEndpoint,
                           version: int,
                           discovered_as: typing.Literal["bidi", "request_response"]) -> typing.Tuple[typing.Dict[str, str] ,str]:
    """
    return restate's discovery object as JSON 
    """
    if version != 1:
        raise ValueError(f"Unsupported protocol version {version}")

    ep = compute_discovery(endpoint, discovered_as)
    json_str = json.dumps(ep, cls=PythonClassEncoder, allow_nan=False)
    headers = {"content-type": "application/vnd.restate.endpointmanifest.v1+json"}
    return (headers, json_str)

def compute_discovery(endpoint: RestateEndpoint, discovered_as : typing.Literal["bidi", "request_response"]) -> Endpoint:
    """
    return restate's discovery object for an endpoint
    """
    services: typing.List[Service] = []

    for service in endpoint.services.values():
        service_type = SERVICE_TYPES[service.service_tag.kind]
        service_handlers = []
        for handler in service.handlers.values():
            # type
            if handler.kind:
                ty = HANDLER_TYPES[handler.kind]
            else:
                ty = None
            # input
            inp = InputPayload(required=False,
                               contentType=handler.handler_io.accept,
                               jsonSchema=None)
            # output
            out = OutputPayload(setContentTypeIfEmpty=False,
                                contentType=handler.handler_io.content_type,
                                jsonSchema=None)
            # add the handler
            service_handlers.append(Handler(name=handler.name, ty=ty, input=inp, output=out))

        # add the service
        services.append(Service(name=service.name, ty=service_type, handlers=service_handlers))

    if endpoint.protocol:
        protocol_mode = PROTOCOL_MODES[endpoint.protocol]
    else:
        protocol_mode = PROTOCOL_MODES[discovered_as]
    return Endpoint(protocolMode=protocol_mode,
                    minProtocolVersion=1,
                    maxProtocolVersion=1,
                    services=services)
